/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


// Configuration reader implementation
// Author: Martin Skogevall, Lumai

#ifndef MAMMUTCONFIG_H
#define MAMMUTCONFIG_H

#include <fstream>
#include <sstream>
#include <vector>
#include <map>

class Configuration
{
private:
	typedef std::map<std::string, std::string> Map;
public:
	std::vector<std::string> ParseCommandLine(std::string const& file);
	void LoadKeyword(std::string const& key, std::string const& value)
	{
		m_map[key] = value;
	}

private:
	std::string Expand(std::string line);
	std::string ExtractKeyword(std::string & str, std::string::size_type first, std::string::size_type & last);

private:
	Map m_map;	
};

std::vector<std::string> Configuration::ParseCommandLine(std::string const& cl)
{
	using namespace std;

	string line = Expand(cl);
	stringstream ss(line);
	vector<string> args;
	string arg;

	while(ss.good())
	{
		ss >> arg;
		args.push_back(arg);
	}
	return args;
}

std::string Configuration::ExtractKeyword(std::string & str, std::string::size_type first, std::string::size_type & last)
{
	using namespace std;
	if (*(str.begin() + first) == '{')
	{
		last = str.find('}', first+1);
		if (last == string::npos)
			throw runtime_error("Keyword was started with '{' but not ended with '}' at position " + first + '.');

		string keyword(str.begin() + first + 1, str.begin() + last);
		return keyword;
	}
	else
		throw runtime_error("You must enclose keywords within '{}'");
}

// Expand all keywords in a single commandline from the configuration
std::string Configuration::Expand(std::string line)
{
	using namespace std;

	string res;
	string::size_type first = 0;
	string::size_type last = line.find('$');
	for(; last != string::npos; last = line.find('$', first))
	{
		if (*(line.begin() + last - 1) == '\\')
		{
			res.append(line.begin() + first, line.begin() + last - 1);
			res.append("$");
		}
		else
		{
			res.append(line.begin() + first, line.begin() + last);

			string keyword = ExtractKeyword(line, last + 1, last);
			Map::const_iterator iter = m_map.find(keyword);
			if (iter != m_map.end())
			{
				res.append(iter->second);
			}
		}
		first = last+1;
	}
	res.append(line.begin() + first, line.end());
	return res;
}

#endif // MAMMUTCONFIG_H
