/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


/* Match rules with nonterminals

set_derives finds, for each variable (nonterminal), which rules can derive it.
It sets up the value of derives so that derives[i - ntokens] points to a
vector of rule numbers, terminated with -1.  */

#include <stdio.h>
#include "system.h"
#include "new.h"
#include "types.h"
#include "gram.h"

short **derives;

void set_derives()
{	
	shorts **dset = NEW2(nvars, shorts *) - ntokens;
	shorts *delts = NEW2(nrules + 1, shorts);	

	shorts *p = delts;
	int i;
	for (i = nrules; i > 0; i--)
    {
		int const lhs = rlhs[i];
		if (lhs >= 0)
		{
			p->next = dset[lhs];
			p->value = i;
			dset[lhs] = p;
			p++;
		}
    }
	
	derives = NEW2(nvars, short *) - ntokens;
	short *q = NEW2(nvars + nrules, short);
	
	for (i = ntokens; i < nsyms; i++)
    {
		derives[i] = q;
		p = dset[i];
		while (p)
		{
			*q++ = p->value;
			p = p->next;
		}
		*q++ = -1;
    }
	
#ifdef	DEBUG
	print_derives();
#endif
	
	FREE(dset + ntokens);
	FREE(delts);
}

void free_derives()
{
	FREE(derives[ntokens]);
	FREE(derives + ntokens);
}

#ifdef DEBUG
extern char **tags;

void print_derives()
{	
	printf("\n\n\nDERIVES\n\n");
	
	for (int i = ntokens; i < nsyms; i++)
    {
		printf("%s derives", tags[i]);
		for (short *sp = derives[i]; *sp > 0; sp++)
			printf("  %d", *sp);
		putchar('\n');
    }
	
	putchar('\n');
}

#endif
