/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


/* Print information on generated parser */

#include <stdio.h>
#include "system.h"
#include "machine.h"
#include "new.h"
#include "files.h"
#include "gram.h"
#include "state.h"

#include <iostream>
#include <iomanip>

using namespace std;


extern char **tags;
extern int nstates;
extern short *accessing_symbol;
extern core **state_table;
extern shifts **shift_table;
extern errs **err_table;
extern reductions **reduction_table;
extern char *consistent;
extern char any_conflicts;
extern char *conflicts;
extern int final_state;

extern void conflict_log();
extern void verbose_conflict_log();
extern void print_reductions(int state);

void print_token();
void print_state();
void print_core();
void print_actions();
void print_grammar();

void terse()
{
	if (any_conflicts)
    {
		conflict_log();
    }
}

void print_core(int state)
{
	int i;
	int k;
	int rule;
	core *statep;
	short *sp;
	short *sp1;
	
	statep = state_table[state];
	k = statep->nitems;
	
	if (k == 0) return;
	
	for (i = 0; i < k; i++)
    {
		sp1 = sp = ritem + statep->items[i];
		
		while (*sp > 0)
			sp++;
		
		rule = -(*sp);
		clog << "    " << tags[rlhs[rule]] << "  ->  ";
		
		for (sp = ritem + rrhs[rule]; sp < sp1; sp++)
			clog << tags[*sp] << ' ';
		
		clog.put('.');
		
		while (*sp > 0)
		{
			clog << ' ' << tags[*sp];
			sp++;
		}
		
		clog << "   (rule " << rule << ")\n";
    }
	
	clog.put('\n');
}


void print_actions(int state)
{
	int i;
	int k;
	int state1;
	int symbol;
	shifts *shiftp;
	errs *errp;
	reductions *redp;
	int rule;
	
	shiftp = shift_table[state];
	redp = reduction_table[state];
	errp = err_table[state];
	
	if (!shiftp && !redp)
    {
		if (final_state == state)
			clog << "    $default\taccept\n";
		else
			clog << "    NO ACTIONS\n";
		return;
    }
	
	if (shiftp)
    {
		k = shiftp->nshifts;
		
		for (i = 0; i < k; i++)
		{
			if (! shiftp->shifts[i]) continue;
			state1 = shiftp->shifts[i];
			symbol = accessing_symbol[state1];
			/* The following line used to be turned off.  */
			if (ISVAR(symbol)) break;
			if (symbol==0)      /* I.e. strcmp(tags[symbol],"$")==0 */
				clog << "    $   \tgo to state " << state1 << '\n';
			else
				clog << "    " << tags[symbol] << "\tshift, and go to state " << state1 << '\n';
		}
		
		if (i > 0)
			clog.put('\n');
    }
	else
    {
		i = 0;
		k = 0;
    }
	
	if (errp)
    {
		int j, nerrs;
		
		nerrs = errp->nerrs;
		
		for (j = 0; j < nerrs; j++)
		{
			if (! errp->errs[j]) continue;
			symbol = errp->errs[j];
			clog << "    " << tags[symbol] << "\terror (nonassociative)\n";
		}
		
		if (j > 0)
			clog.put('\n');
    }
	
	if (consistent[state] && redp)
    {
		rule = redp->rules[0];
		symbol = rlhs[rule];
		clog << "    $default\treduce using rule " << rule << " (" << tags[symbol] << ")\n\n";
    }
	else if (redp)
		print_reductions(state);
	
	if (i < k)
    {
		for (; i < k; i++)
		{
			if (! shiftp->shifts[i]) continue;
			state1 = shiftp->shifts[i];
			symbol = accessing_symbol[state1];
			clog << "    " << tags[symbol] << "\tgo to state " << state1 << "\n";
		}
		
		clog.put('\n');
    }
}

#define END_TEST(end) \
	if (column + strlen(buffer) > (end))					 \
{ clog << buffer << "\n   "; column = 3; buffer[0] = 0; } \
else

void print_state(int state)
{
	clog << "\n\nstate " << state << "\n\n";
	print_core(state);
	print_actions(state);
}

void print_token(int extnum, int token)
{
	clog << " type " << extnum << " is " << tags[token] << '\n';
}

void print_grammar()
{
	int i, j;
	short* rule;
	char buffer[90];
	int column = 0;
	
	/* rule # : LHS -> RHS */
	clog << "\nGrammar\n";
	for (i = 1; i <= nrules; i++)
		/* Don't print rules disabled in reduce_grammar_tables.  */
		if (rlhs[i] >= 0)
		{
			clog << "rule " << i << ' ' << tags[rlhs[i]] << " ->";
			rule = &ritem[rrhs[i]];
			if (*rule > 0)
				while (*rule > 0)
					clog << ' ' << tags[*rule++];
				else
					clog << "		/* empty */";
				clog.put('\n');
		}
		
		/* TERMINAL (type #) : rule #s terminal is on RHS */
		clog << "\nTerminals, with rules where they appear\n\n";
		clog << tags[0] << " (-1)\n";
		if (translations)
		{
			for (i = 0; i <= max_user_token_number; i++)
				if (token_translations[i] != 2)
				{
					buffer[0] = 0;
					column = strlen (tags[token_translations[i]]);
					clog << tags[token_translations[i]];
					END_TEST (50);
					sprintf (buffer, " (%d)", i);
					
					for (j = 1; j <= nrules; j++)
					{
						for (rule = &ritem[rrhs[j]]; *rule > 0; rule++)
							if (*rule == token_translations[i])
							{
								END_TEST (65);
								sprintf (buffer + strlen(buffer), " %d", j);
								break;
							}
					}
					clog << buffer << '\n';
				}
		}
		else for (i = 1; i < ntokens; i++)
		{
			buffer[0] = 0;
			column = strlen (tags[i]);
			clog << tags[i];
			END_TEST (50);
			sprintf (buffer, " (%d)", i);
			
			for (j = 1; j <= nrules; j++)
			{
				for (rule = &ritem[rrhs[j]]; *rule > 0; rule++)
					if (*rule == i)
					{
						END_TEST (65);
						sprintf (buffer + strlen(buffer), " %d", j);
						break;
					}
			}
			clog << buffer << '\n';
		}
		
		clog << "\nNonterminals, with rules where they appear\n\n";
		for (i = ntokens; i <= nsyms - 1; i++)
		{
			int left_count = 0, right_count = 0;
			
			for (j = 1; j <= nrules; j++)
			{
				if (rlhs[j] == i)
					left_count++;
				for (rule = &ritem[rrhs[j]]; *rule > 0; rule++)
					if (*rule == i)
					{
						right_count++;
						break;
					}
			}
			
			buffer[0] = 0;
			clog << tags[i];
			column = strlen (tags[i]);
			sprintf (buffer, " (%d)", i);
			END_TEST (0);
			
			if (left_count > 0)
			{
				END_TEST (50);
				sprintf (buffer + strlen(buffer), " on left:");
				
				for (j = 1; j <= nrules; j++)
				{
					END_TEST (65);
					if (rlhs[j] == i)
						sprintf (buffer + strlen(buffer), " %d", j);
				}
			}
			
			if (right_count > 0)
			{
				if (left_count > 0)
					sprintf (buffer + strlen(buffer), ",");
				END_TEST (50);
				sprintf (buffer + strlen(buffer), " on right:");
				for (j = 1; j <= nrules; j++)
				{
					for (rule = &ritem[rrhs[j]]; *rule > 0; rule++)
						if (*rule == i)
						{
							END_TEST (65);
							sprintf (buffer + strlen(buffer), " %d", j);
							break;
						}
				}
			}
			clog << buffer << '\n';
		}
}

void verbose()
{
	int i;
	
	if (any_conflicts)
		verbose_conflict_log();
	
	print_grammar();
	
	for (i = 0; i < nstates; i++)
		print_state(i);
}

