/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


// Symbol table manager

#include <stdio.h>
#include <string.h>
#include "system.h"
#include "new.h"
#include "symtab.h"
#include "gram.h"

bucket **symtab;
bucket *firstsymbol;
bucket *lastsymbol;

int hash(char *key)
{
	char *cp = key;
	int k = 0;
	
	while (*cp)
		k = ((k << 1) ^ (*cp++)) & 0x3fff;
	
	return (k % TABSIZE);
}

void tabinit()
{
	symtab = NEW2(TABSIZE, bucket *);
	
	firstsymbol = NULL;
	lastsymbol = NULL;
}

bucket *getsym(char *key)
{
	int hashval = hash(key);
	bucket *bp = symtab[hashval];
	
	bool found = false;
	while (bp != NULL)
    {
		if (strcmp(key, bp->tag) == 0)
		{
			found = true;
			break;
		}
		else
			bp = bp->link;
    }
	
	if (!found)
    {
		nsyms++;
		
		bp = NEW(bucket);
		bp->link = symtab[hashval];
		bp->next = NULL;
		bp->tag = strdup(key);
		bp->_class = SUNKNOWN;
		
		if (firstsymbol == NULL)
			firstsymbol = bp;
		else
			lastsymbol->next = bp;

		lastsymbol = bp;		
		symtab[hashval] = bp;
    }
	
	return bp;
}

void free_symtab()
{
	int i;
	bucket *bp, *bptmp;
	
	for (i = 0; i < TABSIZE; i++)
    {
		bp = symtab[i];
		while (bp)
		{
			bptmp = bp->link;

			// TODO: examine if reference counting is needed and if there are mem leaks here
			// This used to cause crashes because one string can appear more than once.
/*			if (bp->type_name)
			{
				FREE(bp->type_name);
				bp->type_name = 0;
			}
*/
			FREE(bp);
			bp = bptmp;
		}
    }
	FREE(symtab);
}
